import os
import sys
# import keyring
import pytest
import configparser
import getpass
import pathlib

import urllib3.connectionpool

from canvasrobot import CanvasRobot
from urltransform import UrlTransformationRobot

# from pymemcache.client import base


def dropdb(thedb):
    for table_name in thedb.tables():
        thedb[table_name].drop()
    thedb.commit()


@pytest.fixture(scope='session')
def cr():
    """ Fixture to set up the test database with test data """
    try:
        cr = CanvasRobot(is_testing=True)
    except urllib3.connectionpool.MaxRetryError:
        print("No connection made to CanvasRobot.")
    # do something to initialise
    # localDAL handles truncate of the tables if is_testing
    else:
        yield cr
    # do something to teardown


@pytest.fixture(scope='session')
def tr():
    """ Fixture to set up the test database with test data """
    try:

        tr = UrlTransformationRobot(is_testing=True)
    except urllib3.connectionpool.MaxRetryError:
        print("No connection")
    # do something to initialise
    # localDAL handles truncate of the tables if is_testing
    else:
        yield tr
    # do something to teardown


@pytest.fixture(scope='session')
def suspend_capture(pytestconfig):
    class suspend_guard:
        def __init__(self):
            self.capmanager = pytestconfig.pluginmanager.getplugin('capturemanager')

        def __enter__(self):
            self.capmanager.suspend_global_capture(in_=True)

        def __exit__(self, _1, _2, _3):
            self.capmanager.resume_global_capture()

    yield suspend_guard()


@pytest.fixture(scope='session')
def user_input(request):
    """ Asks user to check something manually and answer a question
    """

    def _user_input(msg='', hide_pw=False):
        notification = "\n\n<<\tANSWER NEEDED\t>>\n\n{}:".format(msg)

        # suspend input capture by py.test so user input can be recorded here
        capture_manager = request.config.pluginmanager.getplugin('capturemanager')
        capture_manager.suspend_global_capture(in_=True)

        answer = getpass.getpass(prompt=notification) if hide_pw \
            else input(notification)

        # resume capture after question have been asked
        capture_manager.resume_global_capture()

        # logging.debug("Answer: {}".format(answer))
        return answer

    # return the function that will be called by the test
    return _user_input


@pytest.fixture(scope='session')
def sso_config(suspend_capture):
    """
    create ini file if not already present and open it
    :return: parser object for ini file
    """
    dirs = os.path.split(__file__)[0]
    dirs.split(os.path.sep)[-2]

    sso_config = configparser.ConfigParser()
    home = str(pathlib.Path.home())
    path = os.path.join(home,
                        'sso_config.ini')
    if not os.path.exists(path):
        with suspend_capture:
            user_name = ask_user_input('accountname')
            user_passw = ask_password()
        sso_config['user'] = dict(name=user_name,
                                  password=user_passw)
        with open(path, "w") as config_file:  # NOTE don't use binary mode!
            sso_config.write(config_file)
    sso_config.read(path)
    sso_config.path = path  # monkey patch

    return sso_config


def get_platform():
    platforms = {
        'linux1': 'Linux',
        'linux2': 'Linux',
        'darwin': 'OS X',
        'win32': 'Windows'
    }
    if sys.platform not in platforms:
        return sys.platform

    return platforms[sys.platform]


def ask_user_input(msg=''):
    """ Asks user to check something manually and answer a question
    """
    notification = "\n\n>>>\tANSWER NEEDED\t<<<\n\n{}".format(msg)
    answer = input(notification)

    return answer


def ask_password(msg='Password'):
    """ Asks user to check something manually and answer a question
    """
    notification = "\n\n>>>\tANSWER NEEDED\t<<<\n\n{}".format(msg)
    answer = getpass.getpass(notification)

    return answer



import re
import os
import sys
import pathlib
import logging
import sqlite3
import openpyxl
import webview
from attrs import define
import rich_click as click

from result import Ok, Err, Result, is_ok, is_err
from canvasrobot import CanvasRobot, Field


def replace_and_count(original_string: str,
                      search_string: str,
                      replace_string: str,
                      dryrun=False) -> (str, int):
    count = 0
    processed_string = original_string
    while search_string in processed_string:
        processed_string = processed_string.replace(search_string, replace_string, 1)
        count += 1
    return original_string if dryrun else processed_string, count


def show_result(html: str):

    template = """
    <!DOCTYPE html>
    <html>
    <head>
      <title>Zoekresultaat</title>
    </head>
    <body>
      <h2>Mediasite urls without replacement:</h2>
      <p>Klik op de link(s) om de pagina te openen.</p>
      <hr/>
      {}
      <hr/>
      <button onclick='pywebview.api.close()'>Sluit</button>
    </body>
    </html>
    """

    html_with_ui = template.format(html)

    class Api:
        _window = None

        def set_window(self, window):
            self._window = window

        def close(self):
            self._window.destroy()
            self._window = None

            sys.exit(0)  # needed to prevent hang
            # return count, new_body

    api = Api()
    win = webview.create_window(title="Error report (click button to close)",
                                html=html_with_ui,
                                js_api=api)
    api.set_window(win)
    webview.start()


class ImportExcelError(Exception):
    pass


class UrlTransformationRobot(CanvasRobot):
    con = None
    cr = None
    current_page = None
    current_page_url = None
    transformation_report = ""
    pages_changed = 0
    count_replacements = 0

    def __init__(self, db_folder=None, is_testing=False):
        super().__init__(db_folder=db_folder, is_testing=is_testing)  # pass location of database
        self.add_media_ids_table()
        if self.db(self.db.ids).isempty():
            self.import_ids()

    # Begin database section
    def add_media_ids_table(self):
        self.db.define_table('ids',
                             Field('panopto_id', 'string'),
                             Field('mediasite_id', 'string'))

    def import_ids(self):

        ids_result = self.get_video_ids()
        if is_err(ids_result):
            raise ImportExcelError(ids_result.err_value)

        rows = ids_result.ok_value
        print(rows[0])
        assert rows[0]['PanoptoID'] == '532f98ad-43dc-45b6-8109-aeeb01865f0e', \
            "import error in db"
        for row in rows:
            self.db.ids.insert(mediasite_id=row['MediasiteID'],
                               panopto_id=row['PanoptoID'])
        self.db.commit()

    def get_video_ids(self) -> Result[list[dict[str, str]], str]:
        """read ids table from spreadsheet"""
        xls_path = "databases/redirect_list.xlsx"
        try:
            sh = openpyxl.load_workbook(xls_path).active
            column_names = next(sh.values)[0:]
            # Initialize the list of dictionaries
            rows_as_dicts = []
            # Iterate over the sheet rows (excluding header)
            for row in sh.iter_rows(min_row=2, values_only=True):
                row_dict = {column_names[i]: row[i] for i in range(len(column_names))}
                rows_as_dicts.append(row_dict)
            return Ok(rows_as_dicts)
        except Exception as e:
            msg = f"Error opening exported video ID list({e})"
            return Err(msg)

    def lookup_panopto_id(self, mediasite_id: str) -> str:
        db = self.db  # just sugarcoat
        row = db(db.ids.mediasite_id == mediasite_id).select(db.ids.panopto_id).first()
        return row.panopto_id if row else None

    # End database section

    def mediasite2panopto(self, text: str, dryrun=False) -> (str, bool):
        """
        :param text possibly with mediasite urls
        :param dryrun: if true just statistics, no action
        :returns text with updated urls if panopto id are found in lookup
        replace the ms_id with p_id in the
        https://videocollege.uvt.nl/Mediasite/
        Play/ce152c1602144b80bad5a222b7d4cc731d
        replace by (redirect procedure until dec 2024)
        https://tilburguniversity.cloud.panopto.eu/Panopto/Pages/
        Viewer.aspx?id=221a5d47-84ea-44e1-b826-af52017be85c
        """
        updated = False
        updated_text = text
        count_replacements = 0
        # match each source-url and extract the id into a  list of ms_ids
        matches = re.findall(r'(https://videocollege\.uvt\.nl/Mediasite/Play/([a-z0-9]+))', text)

        # for each ms_id: lookup p_id and construct new target-url
        for match in matches:
            ms_url = match[0]
            ms_id = match[1]
            pn_id = self.lookup_panopto_id(ms_id)
            if pn_id:
                # replace source-url with target-url
                pn_url = PN_URL % pn_id
                updated_text, count = replace_and_count(text, ms_url, pn_url, dryrun=dryrun)
                count_replacements += count
                updated = True
            else:
                # no corresponding panopto id found in dv
                msg = (f"<a href={self.current_page_url}>{self.current_page}</a>"
                       f" has mediasite url {ms_url} which is NOT transformed "
                       f"because the mediasite id is not found in DB.")
                logger.info(msg)
                self.transformation_report += (msg + '<br/>')
        return updated_text, updated, count_replacements

    def transform_pages_in_course(self, course_id: int, dryrun=False):
        """
        Transform the mediasite urls in all pages of the course with this course_id
        :param course_id:
        :param dryrun: if true just statistics
        :return:
        """
        pages = self.course_get_pages(course_id)  # example
        for page in pages:
            if page.body:
                self.current_page_url = page.html_url
                self.current_page = page.title
                new_body, updated, count = self.mediasite2panopto(page.body)
                self.count_replacements += count
                if updated:
                    self.pages_changed += 1
                    if not dryrun:
                        page.edit(wiki_page=dict(body=new_body))
        return


@define
class TestCourse:
    id: int


@click.command()
@click.option("--dryrun", default=True, help="If True only show possible result.")
@click.option("--testing", default=False, help="If True test.")
@click.option("--stop_after", default=1, help="Stop after this many courses.")
def run(dryrun, testing, stop_after):
    tr = UrlTransformationRobot()  # default location db: folder 'databases'

    # dryrun = True
    # testing = False

    pages_changed, count_replacements = 0, 0
    test_courses = (TestCourse(34),)

    courses = test_courses if testing else tr.get_courses_in_account()

    # all_ids = (course.id for course in courses)

    # STOP_AFTER = 10

    index = 0
    for index, course in enumerate(courses):

        if index > stop_after:
            break
        tr.transform_pages_in_course(course.id, dryrun=dryrun)

    tr.transformation_report += (f"<p>{index} courses checked with </p>"
                                 f"<p>{tr.pages_changed} "
                                 f"{'pages would be changed' if dryrun else 'changed pages'},"
                                 f" {tr.count_replacements} urls "
                                 f"{'would be' if dryrun else ''} replaced</p>")

    if tr.transformation_report:
        show_result(tr.transformation_report)


if __name__ == '__main__':
    run()


import builtins

input_values = []
print_values = []


def mock_input(s):
    print_values.append(s)
    return input_values.pop(0)


def mock_input_output_start():
    global input_values, print_values

    input_values = []
    print_values = []

    builtins.input = mock_input
    builtins.print = lambda s: print_values.append(s)


def get_display_output():
    global print_values
    return print_values


def set_keyboard_input(mocked_inputs):
    global input_values

    mock_input_output_start()
    input_values = mocked_inputs



import mock
import builtins
import pytest
import webview
from canvasrobot.canvasrobot import Course2Foldername

"""
1. note that this is real live testing when interfacing with Canvas
2. please adjust the 4 constants below, 
3. at first instance of CanvasRobot you have to supply
   the URL of your Canvas environment, your Canvas API key and 
   the admin_id. 
   Both will be recorded in a secure location using the 
   [keyring](https://pypi.org/project/keyring/) library.
   Use -s as additional argument to allow user input in pytest.
   You can remove it later, after you have given the data.
"""


ADMIN_ID = 6  # If no admin_id available: set to 0
A_TEACHER_ID = 8  # choose a teacher/teacher_id from Canvas
NR_COURSES_TEACHER = 8  # lookup number of courses for this teacher
NR_COURSES_ADMIN = 178  # lookup using canvas website
TEST_COURSE = 34  # first create this test course in Canvas
TEST_COURSE_NR_ASSIGNMENTS = 8  # check if there are this number of assignments
TEST_COURSE_NR_EXAMINATIONS = 3  # make sure there are this many files in folder Tentamens

# - with assignments and at least one graded submission with file upload
# - and at least one file in folder 'Tentamens' in Files


def tst_init(cr):
    """ something with capture"""
    inputs = iter(['https://tilburguniversity.instructure.com', 'a key', 8])
    with mock.patch.object(builtins, 'input', lambda _: next(inputs)):
        assert cr


def test_getcourses_current_teacher(cr):
    """ for current teacher get the courses"""
    courses = cr.get_courses("teacher")
    assert len(list(courses)) == NR_COURSES_TEACHER


def test_getcourses_admin(cr):
    """ for the admin account (if available) get the courses"""
    if ADMIN_ID:
        courses = cr.get_courses_in_account()
        assert len(list(courses)) == NR_COURSES_ADMIN


def test_api_valid(cr):
    """you need -s parameter in pytest once to record API key in keyring"""

    user = cr.get_user(8)
    assert user.id == 8


def test_course_metadata(cr):
    """ test if course_metadata collects the right data from the test course"""
    ignore_examination_names = ["Opdracht 1", ]
    md = cr.course_metadata(TEST_COURSE, ignore_examination_names)
    assert md.assignments_summary, "field assignments_summary not there"
    assert f"graded {TEST_COURSE_NR_ASSIGNMENTS}" in md.assignments_summary, \
        "assignments not reported in assignments summary"
    assert md.examination_records, "field examination_record not there"
    assert f"Total: {TEST_COURSE_NR_EXAMINATIONS}" in md.examinations_summary, \
        "examination files not (complete) in summary"


@pytest.mark.parametrize("course_id,nr_assignments,nr_examination_files",
                         [(TEST_COURSE, TEST_COURSE_NR_ASSIGNMENTS, TEST_COURSE_NR_EXAMINATIONS),
                          (12519, 0, 0)])
def test_update_database_from_canvas(cr,
                                     course_id,
                                     nr_assignments,
                                     nr_examination_files):
    """ To check the proces 'create and update course data':
     first delete the test course from local database if its there,
     then run update course from Canvas source (which should add it again)
     check if it's back and containing the expected data."""
    result = cr.delete_course_from_database(course_id)
    print(f"{result} records (recursive) deleted")
    if result == 0:
        print("Note: update database from canvas could not not be tested")

        return True  # I know, but we cannot update if it's missing!
    cr.update_database_from_canvas(single_course=course_id)

    course = cr.get_course_from_database(course_id)
    assert course.course_id == course_id, "course not found, not added"
    assert course.assignments_summary, "field assignments_summary not there"
    assert f"graded {nr_assignments}" if nr_assignments \
        else "No assignments" in course.assignments_summary, \
        "no of assignments not correctly reported in assignments summary"
    # assert course.examination_candidates, "field examination_candidates not there"
    assert f"Total: {nr_examination_files}" if nr_examination_files \
        else "No examination files" in course.examinations_summary, \
        "# of examination files not correct in summary"


def test_update_record_db(cr):
    """ Check the database after updating single record"""
    cr.update_database_from_canvas(single_course=TEST_COURSE)

    db = cr.db
    qry = (db.course.course_id == TEST_COURSE)
    cr.update_record_db(qry, "examinations_ok", True)
    course = cr.get_course_from_database(TEST_COURSE)
    assert course, f"Course {TEST_COURSE} not found"
    assert course.examinations_ok is True
    cr.update_record_db(qry, "examinations_ok", False)
    course = cr.get_course_from_database(TEST_COURSE)
    assert course.examinations_ok is False


def test_outliner_foldernames(cr):
    cr.outliner_foldernames.append(Course2Foldername(34, "Tentamens"))
    cr.outliner_foldernames.append(Course2Foldername(35, "Tentamens 3"))

    output = cr.print_outliner_foldernames()

    assert "34" in output, f"error in printing foldernames: {output}"
    assert "35" in output, f"error in printing foldernames: {output}"
    assert "Tentamens" in output, f"error in printing foldernames: {output}"
    assert "Tentamens 3" in output, f"error in printing foldernames: {output}"


def test_search_replace(cr):
    """check if course_search_replace function works"""
    db = cr.db
    course = cr.get_course(TEST_COURSE)
    pages = course.get_pages(include=['body'])
    search_text, replace_text = ' je', ' u'
    page_found_url = ""
    dryrun = True
    for page in pages:
        if search_text in page.body:
            page_found_url = page.url  # remember
            cr.search_replace_in_page(page, search_text, replace_text, dryrun=dryrun)
            # We only need one page to test this
            break

    if page_found_url:
        if not dryrun:
            # read again from canvas instance to check
            page = course.get_page(page_found_url)
            assert search_text not in page.body
            assert replace_text in page.body
    else:
        assert False, f"Source string '{search_text}' not found in any page of course {TEST_COURSE}"



from typing import List
from datetime import datetime
from attrs import define
# from dataclasses import dataclass, field


@define
class Course:
    course_id: int = 0
    name: str = ""
    course_code: str = ""
    sis_code: str = ""
    creation_date: datetime = None
    ac_year: int = 1961
    teachers: List[str] = []
    teacher_names: List[str] = []


@define
class EnrollDTO:
    username: str
    course: str
    role: str
    user_id: int = 0
    course_id: int = 0

@define
class SearchTextInCourseDTO:
    course_id: int = 0
    course: str = ""
    search: str = ""

@define
class ExaminationDTO:
    course_id: int
    course_name: str
    name:str

@define
class CourseMetadata:
    nr_modules: int
    nr_module_items: int
    nr_pages: int
    nr_assignments: int
    nr_quizzes: int
    nr_files: int
    assignments_summary: str
    examinations_summary: str
    examination_records: list[ExaminationDTO]
    # nr_collaborations: int

    # #ext_urls: int
    # avr_len_assignments: int

@define
class Grade:
    stud_name: str
    stud_id: str
    final_score: float
    final_grade: float





from .course import Course, EnrollDTO, SearchTextInCourseDTO, \
    CourseMetadata, Grade, ExaminationDTO
from .user import User
from .guest import Guest
from .quiz import Answer, QuizDTO, QuestionDTO, Stats

__all__ = ["Course","EnrollDTO","SearchTextInCourseDTO",
           "CourseMetadata","Grade","ExaminationDTO",
           "User","Guest",
           "Answer","QuizDTO","QuestionDTO","Stats"]




import pytest
import textwrap


def test_mediasite2panopto(tr):
    """
    :param tr: fixture: the TransformationRobot based on CanvasRobot
    :returns: True if url is transformed and 'bad' url is not transformed and reported
    """

    source = textwrap.dedent("""\
    replace the ms_id with p_id in the

    https://videocollege.uvt.nl/Mediasite/Play/ce152c1602144b80bad5a222b7d4cc731d]

    Nu een link met id die niet bestaat https://videocollege.uvt.nl/Mediasite/Play/ce152c1602144b80bad5a222b7d4cc731 
    is dus niet goed 

    replace by (redirect procedure until dec 2024)


    """)
    target, updated = tr.mediasite2panopto(source)
    print(target)

    assert updated, "'updated' should be 'True' as 'source' contains a videocollege url"
    assert ('https://tilburguniversity.cloud.panopto.eu/Panopto/'
            'Pages/Viewer.aspx?id=221a5d47-84ea-44e1-b826-af52017be85c') in target
    # don't change non-redirecting urls, report them
    bad_ms_url = 'https://videocollege.uvt.nl/Mediasite/Play/ce152c1602144b80bad5a222b7d4cc731'
    assert bad_ms_url in target, f"{bad_ms_url} should not be changed"

    assert bad_ms_url in tr.transformation_report



from .canvasrobot import CanvasRobot, LocalDAL, \
    EDUCATIONS, COMMUNITIES
from .canvasrobot_model import STUDADMIN, SHORTNAMES, Field
from .urltransform import UrlTransformationRobot, show_result
from .commandline import show_search_result, search_replace_show, get_logger

__all__ = ["CanvasRobot", "UrlTransformationRobot", "LocalDAL", "Field",
           "ENROLLMENT_TYPES", "EDUCATIONS", "COMMUNITIES",
           "STUDADMIN", "SHORTNAMES",
           "get_logger", "show_result", "show_search_result", "search_replace_show"]

__version__ = "0.8.4"  # It MUST match the version in pyproject.toml file
